@@ -2,8 +2,6 @@
 #include "d3d11_context.h"
 #include "d3d11_device.h"
 
-#include "../dxvk/dxvk_data.h"
-
 namespace dxvk {
   
   D3D11Buffer::D3D11Buffer(
@@ -84,6 +82,10 @@ namespace dxvk {
         info.access |= VK_ACCESS_HOST_WRITE_BIT;
     }
 
+    // Always enable BDA usage if available so that CUDA interop can work
+    if (m_parent->GetDXVKDevice()->features().vk12.bufferDeviceAddress)
+      info.usage |= VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT;
+
     if (p11on12Info) {
       m_11on12 = *p11on12Info;
 
@@ -94,32 +96,25 @@ namespace dxvk {
       if (m_desc.CPUAccessFlags)
         m_11on12.Resource->Map(0, nullptr, &importInfo.mapPtr);
 
-      // D3D12 will always have BDA enabled
-      info.usage |= VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT;
-
       m_buffer = m_parent->GetDXVKDevice()->importBuffer(info, importInfo, GetMemoryFlags());
-      m_mapped = m_buffer->getSliceHandle();
-
+      m_cookie = m_buffer->cookie();
+      m_mapPtr = m_buffer->mapPtr(0);
       m_mapMode = DetermineMapMode(m_buffer->memFlags());
     } else if (!(pDesc->MiscFlags & D3D11_RESOURCE_MISC_TILE_POOL)) {
       VkMemoryPropertyFlags memoryFlags = GetMemoryFlags();
       m_mapMode = DetermineMapMode(memoryFlags);
 
-      // Prevent buffer renaming for buffers that are not mappable, so
-      // that interop interfaces can safely query the GPU address.
-      if (m_mapMode == D3D11_COMMON_BUFFER_MAP_MODE_NONE
-       && m_parent->GetDXVKDevice()->features().vk12.bufferDeviceAddress)
-        info.usage |= VK_BUFFER_USAGE_SHADER_DEVICE_ADDRESS_BIT;
-
       // Create the buffer and set the entire buffer slice as mapped,
       // so that we only have to update it when invalidating the buffer
       m_buffer = m_parent->GetDXVKDevice()->createBuffer(info, memoryFlags);
-      m_mapped = m_buffer->getSliceHandle();
+      m_cookie = m_buffer->cookie();
+      m_mapPtr = m_buffer->mapPtr(0);
     } else {
       m_sparseAllocator = m_parent->GetDXVKDevice()->createSparsePageAllocator();
       m_sparseAllocator->setCapacity(info.size / SparseMemoryPageSize);
 
-      m_mapped = DxvkBufferSliceHandle();
+      m_cookie = 0u;
+      m_mapPtr = nullptr;
       m_mapMode = D3D11_COMMON_BUFFER_MAP_MODE_NONE;
     }
 
