@@ -48,6 +48,9 @@ namespace dxvk {
     bool                IsBackBuffer;
     bool                IsAttachmentOnly;
     bool                IsLockable;
+
+    // Additional parameters for ID3D9VkInteropDevice
+    VkImageUsageFlags   ImageUsage = 0;
   };
 
   struct D3D9ColorView {
@@ -72,6 +75,8 @@ namespace dxvk {
 
   public:
 
+    static constexpr UINT AllLayers = UINT32_MAX;
+
     D3D9CommonTexture(
             D3D9DeviceEx*             pDevice,
             IUnknown*                 pInterface,
@@ -334,10 +339,6 @@ namespace dxvk {
 
     void MarkAllNeedReadback() { m_needsReadback.setAll(); }
 
-    void SetReadOnlyLocked(UINT Subresource, bool readOnly) { return m_readOnly.set(Subresource, readOnly); }
-
-    bool GetReadOnlyLocked(UINT Subresource) const { return m_readOnly.get(Subresource); }
-
     const Rc<DxvkImageView>& GetSampleView(bool srgb) const {
       return m_sampleView.Pick(srgb && IsSrgbCompatible());
     }
@@ -519,8 +520,6 @@ namespace dxvk {
 
     D3D9SubresourceBitset         m_locked = { };
 
-    D3D9SubresourceBitset         m_readOnly = { };
-
     D3D9SubresourceBitset         m_needsReadback = { };
 
     D3D9SubresourceBitset         m_needsUpload = { };
@@ -547,11 +546,6 @@ namespace dxvk {
       const DxvkImageCreateInfo*  pImageInfo,
             VkImageTiling         Tiling) const;
 
-    VkImageUsageFlags EnableMetaCopyUsage(
-            VkFormat              Format,
-            VkImageTiling         Tiling,
-            VkSampleCountFlags    SampleCount) const;
-
     D3D9_COMMON_TEXTURE_MAP_MODE DetermineMapMode() const;
 
     VkImageLayout OptimizeLayout(
@@ -563,8 +557,6 @@ namespace dxvk {
             D3DRESOURCETYPE  Dimension,
             UINT             Layer);
 
-    static constexpr UINT AllLayers = UINT32_MAX;
-
   };
 
 }
