@@ -9,35 +9,41 @@ namespace dxvk {
     DxvkOptions(const Config& config);
 
     /// Enable debug utils
-    bool enableDebugUtils;
+    bool enableDebugUtils = false;
 
     /// Enable state cache
-    bool enableStateCache;
+    bool enableStateCache = true;
+
+    /// Enable memory defragmentation
+    Tristate enableMemoryDefrag = Tristate::Auto;
 
     /// Number of compiler threads
     /// when using the state cache
-    int32_t numCompilerThreads;
+    int32_t numCompilerThreads = 0;
 
     /// Enable graphics pipeline library
-    Tristate enableGraphicsPipelineLibrary;
+    Tristate enableGraphicsPipelineLibrary = Tristate::Auto;
 
     /// Enables pipeline lifetime tracking
-    Tristate trackPipelineLifetime;
+    Tristate trackPipelineLifetime = Tristate::Auto;
 
     /// Shader-related options
-    Tristate useRawSsbo;
+    Tristate useRawSsbo = Tristate::Auto;
 
     /// HUD elements
     std::string hud;
 
     /// Forces swap chain into MAILBOX (if true)
     /// or FIFO_RELAXED (if false) present mode
-    Tristate tearFree;
+    Tristate tearFree = Tristate::Auto;
 
     // Hides integrated GPUs if dedicated GPUs are
     // present. May be necessary for some games that
     // incorrectly assume monitor layouts.
-    bool hideIntegratedGraphics;
+    bool hideIntegratedGraphics = false;
+
+    /// Clears all mapped memory to zero.
+    bool zeroMappedMemory = false;
 
     // Device name
     std::string deviceFilter;
