@@ -1,9 +1,10 @@
 #pragma once
 
+#include <atomic>
 #include <map>
 
+#include "dxvk_access.h"
 #include "dxvk_memory.h"
-#include "dxvk_resource.h"
 
 namespace dxvk {
 
@@ -18,16 +19,6 @@ namespace dxvk {
 
   constexpr static VkDeviceSize SparseMemoryPageSize = 1ull << 16;
 
-  /**
-   * \brief Sparse page handle
-   */
-  struct DxvkSparsePageHandle {
-    VkDeviceMemory  memory;
-    VkDeviceSize    offset;
-    VkDeviceSize    length;
-  };
-
-
   /**
    * \brief Buffer info for sparse page
    *
@@ -165,38 +156,6 @@ namespace dxvk {
   };
 
 
-  /**
-   * \brief Sparse memory page
-   *
-   * Stores a single reference-counted page
-   * of memory. The page size is 64k.
-   */
-  class DxvkSparsePage : public DxvkResource {
-
-  public:
-
-    DxvkSparsePage(DxvkMemory&& memory)
-    : m_memory(std::move(memory)) { }
-
-    /**
-     * \brief Queries memory handle
-     * \returns Memory information
-     */
-    DxvkSparsePageHandle getHandle() const {
-      DxvkSparsePageHandle result;
-      result.memory = m_memory.memory();
-      result.offset = m_memory.offset();
-      result.length = m_memory.length();
-      return result;
-    }
-
-  private:
-
-    DxvkMemory  m_memory;
-
-  };
-
-
   /**
    * \brief Sparse page mapping
    *
@@ -223,11 +182,11 @@ namespace dxvk {
      * \brief Queries memory handle
      * \returns Memory information
      */
-    DxvkSparsePageHandle getHandle() const {
-      if (m_page == nullptr)
-        return DxvkSparsePageHandle();
+    DxvkResourceMemoryInfo getMemoryInfo() const {
+      if (!m_page)
+        return DxvkResourceMemoryInfo();
 
-      return m_page->getHandle();
+      return m_page->getMemoryInfo();
     }
 
     bool operator == (const DxvkSparseMapping& other) const {
@@ -239,18 +198,18 @@ namespace dxvk {
       return m_page != other.m_page;
     }
 
-    operator bool () const {
+    explicit operator bool () const {
       return m_page != nullptr;
     }
 
   private:
 
     Rc<DxvkSparsePageAllocator> m_pool;
-    Rc<DxvkSparsePage>          m_page;
+    Rc<DxvkResourceAllocation>  m_page;
 
     DxvkSparseMapping(
             Rc<DxvkSparsePageAllocator> allocator,
-            Rc<DxvkSparsePage>          page);
+            Rc<DxvkResourceAllocation>  page);
 
     void acquire() const;
 
@@ -304,15 +263,13 @@ namespace dxvk {
     dxvk::mutex                       m_mutex;
     uint32_t                          m_pageCount = 0u;
     uint32_t                          m_useCount = 0u;
-    std::vector<Rc<DxvkSparsePage>>   m_pages;
-
-    Rc<DxvkSparsePage> allocPage();
+    std::vector<Rc<DxvkResourceAllocation>> m_pages;
 
     void acquirePage(
-      const Rc<DxvkSparsePage>&   page);
+      const Rc<DxvkResourceAllocation>& page);
 
     void releasePage(
-      const Rc<DxvkSparsePage>&   page);
+      const Rc<DxvkResourceAllocation>& page);
 
   };
 
@@ -331,17 +288,19 @@ namespace dxvk {
 
     DxvkSparsePageTable(
             DxvkDevice*             device,
-      const DxvkBuffer*             buffer);
+      const VkBufferCreateInfo&     bufferInfo,
+            VkBuffer                bufferHandle);
 
     DxvkSparsePageTable(
             DxvkDevice*             device,
-      const DxvkImage*              image);
+      const VkImageCreateInfo&      imageInfo,
+            VkImage                 imageHandle);
 
     /**
      * \brief Checks whether page table is defined
      * \returns \c true if the page table is defined
      */
-    operator bool () const {
+    explicit operator bool () const {
       return m_buffer || m_image;
     }
 
@@ -455,8 +414,8 @@ namespace dxvk {
 
   private:
 
-    const DxvkBuffer* m_buffer  = nullptr;
-    const DxvkImage*  m_image   = nullptr;
+    VkBuffer m_buffer = VK_NULL_HANDLE;
+    VkImage m_image = VK_NULL_HANDLE;
 
     DxvkSparseImageProperties                         m_properties    = { };
     std::vector<DxvkSparseImageSubresourceProperties> m_subresources;
@@ -469,26 +428,224 @@ namespace dxvk {
   /**
    * \brief Paged resource
    *
-   * Base class for any resource that can
-   * hold a sparse page table.
+   * Base class for memory-backed resources that may
+   * or may not also have a sparse page table.
    */
-  class DxvkPagedResource : public DxvkResource {
+  class DxvkPagedResource {
 
   public:
 
+    DxvkPagedResource()
+    : m_cookie(++s_cookie) { }
+
+    virtual ~DxvkPagedResource();
+
     /**
-     * \brief Queries sparse page table
-     * \returns Sparse page table, if defined
+     * \brief Queries resource cookie
+     * \returns Resource cookie
      */
-    DxvkSparsePageTable* getSparsePageTable() {
-      return m_sparsePageTable
-        ? &m_sparsePageTable
-        : nullptr;
+    uint64_t cookie() const {
+      return m_cookie;
     }
 
-  protected:
+    /**
+     * \brief Increments reference count
+     */
+    force_inline void incRef() {
+      acquire(DxvkAccess::None);
+    }
 
-    DxvkSparsePageTable m_sparsePageTable;
+    /**
+     * \brief Decrements reference count
+     */
+    force_inline void decRef() {
+      release(DxvkAccess::None);
+    }
+
+    /**
+     * \brief Acquires resource with given access
+     *
+     * Atomically increments both the reference count
+     * as well as the use count for the given access.
+     */
+    force_inline void acquire(DxvkAccess access) {
+      m_useCount.fetch_add(getIncrement(access), std::memory_order_acquire);
+    }
+
+    /**
+     * \brief Releases resource with given access
+     *
+     * Atomically decrements both the reference count
+     * as well as the use count for the given access.
+     */
+    force_inline void release(DxvkAccess access) {
+      uint64_t increment = getIncrement(access);
+      uint64_t remaining = m_useCount.fetch_sub(increment, std::memory_order_release);
+
+      if (unlikely(remaining == increment))
+        delete this;
+    }
+
+    /**
+     * \brief Converts reference type
+     *
+     * \param [in] from Old access type
+     * \param [in] to New access type
+     */
+    force_inline void convertRef(DxvkAccess from, DxvkAccess to) {
+      uint64_t increment = getIncrement(to) - getIncrement(from);
+
+      if (increment)
+        m_useCount.fetch_add(increment, std::memory_order_acq_rel);
+    }
+
+    /**
+     * \brief Checks whether resource is in use
+     * 
+     * Returns \c true if there are pending accesses to
+     * the resource by the GPU matching the given access
+     * type. Note that checking for reads will also return
+     * \c true if the resource is being written to.
+     * \param [in] access Access type to check for
+     * \returns \c true if the resource is in use
+     */
+    force_inline bool isInUse(DxvkAccess access) const {
+      return m_useCount.load(std::memory_order_acquire) >= getIncrement(access);
+    }
+
+    /**
+     * \brief Tries to acquire reference
+     *
+     * If the reference count is zero at the time this is called,
+     * the method will fail, otherwise the reference count will
+     * be incremented by one. This is useful to safely obtain a
+     * pointer from a look-up table that does not own references.
+     * \returns \c true on success
+     */
+    Rc<DxvkPagedResource> tryAcquire() {
+      uint64_t increment = getIncrement(DxvkAccess::None);
+      uint64_t refCount = m_useCount.load(std::memory_order_acquire);
+
+      do {
+        if (!refCount)
+          return nullptr;
+      } while (!m_useCount.compare_exchange_strong( refCount,
+        refCount + increment, std::memory_order_relaxed));
+
+      return Rc<DxvkPagedResource>::unsafeCreate(this);
+    }
+
+    /**
+     * \brief Sets tracked command list ID
+     *
+     * Used to work out if a resource has been used in the current
+     * command list and optimize certain transfer operations.
+     * \param [in] trackingId Tracking ID
+     * \param [in] access Tracked access
+     * \returns \c true if the tracking ID was updated, or \c false
+     *    if the resource was already tracked with the same ID.
+     */
+    bool trackId(uint64_t trackingId, DxvkAccess access) {
+      // Encode write access in the least significant bit
+      uint64_t trackId = (trackingId << 1u) + uint64_t(access == DxvkAccess::Write);
+
+      if (trackId <= m_trackId)
+        return false;
+
+      m_trackId = trackId;
+      return true;
+    }
+
+    /**
+     * \brief Checks whether a resource has been tracked
+     *
+     * \param [in] trackingId Current tracking ID
+     * \param [in] access Destination access
+     * \returns \c true if the resource has been used in a way that
+     *    prevents recordering commands with the given resource access.
+     */
+    bool isTracked(uint64_t trackingId, DxvkAccess access) const {
+      // We actually want to check for read access here so that this check only
+      // fails if the resource hasn't been used or if both accesses are read-only.
+      return m_trackId >= (trackingId << 1u) + uint64_t(access != DxvkAccess::Write);
+    }
+
+    /**
+     * \brief Resets tracking
+     *
+     * Marks the resource as unused in the current command list.
+     * Should be done when assigning new backing storage.
+     */
+    void resetTracking() {
+      m_trackId = 0u;
+    }
+
+    /**
+     * \brief Queries sparse page table
+     *
+     * Should be removed once storage objects can
+     * be retrieved from resources diectly.
+     * \returns Sparse page table, if defined
+     */
+    virtual DxvkSparsePageTable* getSparsePageTable() = 0;
+
+    /**
+     * \brief Allocates new backing storage with constraints
+     *
+     * \param [in] mode Allocation mode flags to control behaviour.
+     *    When relocating the resource to a preferred memory type,
+     *    \c NoFallback should be set, when defragmenting device
+     *    memory then \c NoAllocation should also be set.
+     * \returns \c true in the first field if the operation is
+     *    considered successful, i.e. if an new backing allocation
+     *    was successfully created or is unnecessary. The second
+     *    field will contain the new allocation itself.
+     */
+    virtual Rc<DxvkResourceAllocation> relocateStorage(
+            DxvkAllocationModes         mode) = 0;
+
+  private:
+
+    std::atomic<uint64_t> m_useCount = { 0u };
+    uint64_t              m_trackId = { 0u };
+    uint64_t              m_cookie = { 0u };
+
+    static constexpr uint64_t getIncrement(DxvkAccess access) {
+      return uint64_t(1u) << (uint32_t(access) * 20u);
+    }
+
+    static std::atomic<uint64_t> s_cookie;
+
+  };
+
+
+  /**
+   * \brief Typed tracking reference for resources
+   *
+   * Does not provide any access information.
+   */
+  class DxvkResourceRef : public DxvkTrackingRef {
+    constexpr static uintptr_t AccessMask = 0x3u;
+
+    static_assert(alignof(DxvkPagedResource) > AccessMask);
+  public:
+
+    template<typename T>
+    explicit DxvkResourceRef(Rc<T>&& object, DxvkAccess access)
+    : m_ptr(reinterpret_cast<uintptr_t>(static_cast<DxvkPagedResource*>(object.ptr())) | uintptr_t(access)) {
+      object.unsafeExtract()->convertRef(DxvkAccess::None, access);
+    }
+
+    explicit DxvkResourceRef(DxvkPagedResource* object, DxvkAccess access)
+    : m_ptr(reinterpret_cast<uintptr_t>(object) | uintptr_t(access)) {
+      object->acquire(access);
+    }
+
+    ~DxvkResourceRef();
+
+  private:
+
+    uintptr_t m_ptr = 0u;
 
   };
 
@@ -673,7 +830,7 @@ namespace dxvk {
      */
     void bindBufferMemory(
       const DxvkSparseBufferBindKey& key,
-      const DxvkSparsePageHandle&   memory);
+      const DxvkResourceMemoryInfo&  memory);
 
     /**
      * \brief Adds an image memory bind
@@ -683,7 +840,7 @@ namespace dxvk {
      */
     void bindImageMemory(
       const DxvkSparseImageBindKey& key,
-      const DxvkSparsePageHandle&   memory);
+      const DxvkResourceMemoryInfo& memory);
 
     /**
      * \brief Adds an opaque image memory bind
@@ -693,7 +850,7 @@ namespace dxvk {
      */
     void bindImageOpaqueMemory(
       const DxvkSparseImageOpaqueBindKey& key,
-      const DxvkSparsePageHandle&   memory);
+      const DxvkResourceMemoryInfo& memory);
 
     /**
      * \brief Submits sparse binding operation
@@ -723,17 +880,17 @@ namespace dxvk {
     std::vector<uint64_t>     m_signalSemaphoreValues;
     std::vector<VkSemaphore>  m_signalSemaphores;
 
-    std::map<DxvkSparseBufferBindKey,      DxvkSparsePageHandle> m_bufferBinds;
-    std::map<DxvkSparseImageBindKey,       DxvkSparsePageHandle> m_imageBinds;
-    std::map<DxvkSparseImageOpaqueBindKey, DxvkSparsePageHandle> m_imageOpaqueBinds;
+    std::map<DxvkSparseBufferBindKey,      DxvkResourceMemoryInfo> m_bufferBinds;
+    std::map<DxvkSparseImageBindKey,       DxvkResourceMemoryInfo> m_imageBinds;
+    std::map<DxvkSparseImageOpaqueBindKey, DxvkResourceMemoryInfo> m_imageOpaqueBinds;
 
     static bool tryMergeMemoryBind(
             VkSparseMemoryBind&               oldBind,
       const VkSparseMemoryBind&               newBind);
 
     static bool tryMergeImageBind(
-            std::pair<DxvkSparseImageBindKey, DxvkSparsePageHandle>& oldBind,
-      const std::pair<DxvkSparseImageBindKey, DxvkSparsePageHandle>& newBind);
+            std::pair<DxvkSparseImageBindKey, DxvkResourceMemoryInfo>& oldBind,
+      const std::pair<DxvkSparseImageBindKey, DxvkResourceMemoryInfo>& newBind);
 
     void processBufferBinds(
             DxvkSparseBufferBindArrays&       buffer);
